""" Common utilities for the plugin and server backends. """
import sys
import hashlib
import os
import struct
import termios
import fcntl
from urllib.request import urlretrieve
from math import ceil


def levenshtein(first, second):
    """ Get the levenshtein edit distance between the strings.

    The Levenshtein distance is how many characters that need to be changed,
    deleted or inserted in the first string to form the second string. It is
    used to determine the equality of two strings.

    This function accepts to parameters:
        first     The first string.
        second    The second string.

    """
    if len(first) < len(second):
        return levenshtein(second, first)

    if len(second) == 0:
        return len(first)

    previous_row = range(len(second) + 1)

    for i, k in enumerate(first):
        current_row = [i + 1]

        for index, value in enumerate(second):
            insertions = previous_row[index + 1] + 1
            deletions = current_row[index] + 1
            substitutions = previous_row[index] + (k != value)

            current_row.append(min(insertions, deletions, substitutions))

        previous_row = current_row

    return previous_row[-1]


def list_names(array, separator=', ', last_separator=' and '):
    """ Return a string with a listing of the elements.

    This function will basically first join the elements in the list by the
    parameter 'separator'(defaults to ', '), but the last field will be
    appended with the parameter 'last_separator'(defaults to ' and ') before
    it. This leads to a nice human readable representation of the list.

    Examples:
    >>> list_names(['Jane', 'Joe', 'Peter'])
    "Jane, Joe and Peter"
    >>> list_names(['Joe', 'Jane'])
    "Joe and Jane"
    >>> list_names(['Jane'])
    "Jane"

    The function accepts up to three parameters:
        array                    The list to represent.
        separator=', '           The separator between the first elements in
                                 the list.
        last_separator=' and '   The separator for the last element.

    """
    if len(array) > 1:
        return ''.join([separator.join([str(i) for i in array[:-1]]),
                        last_separator,
                        str(array[-1])])
    elif len(array) == 1:
        return array[0]


def download(url, destination=None, checksum=None, prefix='',
             display_name=None):
    """ Download with progressbar.

    Arguments:
        url             URL to download from.
        destination     Path to download the file to. Defaults to None. When it
                        is none the destination is based on the url, and the
                        folder is the CWD.
        checksum        MD5 checksum to check the file against. Defaults to
                        None. When it is none, no checksum validation is done.
        prefix          What to put before the file_name on the left side.
                        Defaults to an empty string.
        display_name    The name to display on the left side instead of the
                        destination. Defaults to None.

    """
    if destination is None:
        destination = url.split('/')[-1]
    if display_name is None:
        display_name = destination

    term_width = get_term_width()
    pprefix = prefix + display_name
    progress = create_progress_bar(prefix=pprefix, width=term_width)
    urlretrieve(url, filename=destination, reporthook=progress)

    if checksum is not None and len(checksum) > 0:
        print('\n' + ' ' * len(prefix) + 'Checking checksum...', end=' ')
        actual_checksum = checksum_file(destination)
        if actual_checksum == checksum:
            print('Sucess')
        else:
            os.remove(destination)
            print('The checksums did not match! The file was deleted.')


def create_progress_bar(width, prefix=None):
    """ Create a progress bar.

    This returns a function which can be used as the progress hook for
    urlretrieve. If you don't want to use the function of urlretrieve, the
    signature for the function is int, int, int: Count, count_size, max.

    Intended for the download function.

    """
    left = ceil(width/2)
    right = width-left
    net_bar_size = right-8
    char_size = net_bar_size/100
    frmt = '{{prefix:<{}}}[{{bar_left}}{{bar_right}}] {{prc}}%\r'.format(left)
    last_percent = -1

    def progress_hook(count, blocksize, totalsize):
        """ Progress hook. """
        percent = min(ceil(count*blocksize*100/totalsize), 100)
        nonlocal last_percent
        if percent == last_percent:
            return
        else:
            last_percent = percent
        fill_size = int(char_size*percent)
        empty_size = net_bar_size-fill_size
        text = frmt.format(prefix=prefix,
                           bar_left=('='*(fill_size-1)+'>' if percent < 100
                                     else '='*fill_size),
                           bar_right=' '*empty_size,
                           prc=percent)

        sys.stdout.write(text)
        sys.stdout.flush()
    return progress_hook


def checksum_file(file):
    """ MD5 Checksum of file.

    This function will return the MD5 checksum of the file.

    One parameter is accepted:
        file    The name of the file to checksum, or the (relative) path to it.

    """
    if type(file) is str:
        with open(file, 'rb') as file:
            return checksum_file(file)
    else:
        return hashlib.md5(file.read()).hexdigest()


def replace_last(string, old, new):
    """ Replace the last occurance of `old` in `string` with `new`. """
    return new.join(string.rsplit(old, 1))


def get_term_width(term=1):
    """ Get the width of the terminal.

    This tries to exctract the width with fcntl.ioctl. If that failes the
    standard width 80 is returned.

    This function is using the fcntl and ioctl system calls which are only
    available on Unix.

    """
    try:
        return struct.unpack('hh', fcntl.ioctl(term,
                                               termios.TIOCGWINSZ,
                                               '1234'))[1]
    except BaseException:
        return 80


def extract_file(zipped, file, dest):
    """ Extract file from ZipFile archive to file.

    This function will extract the file in `file` in the zip archive to the
    file in `dest` in the filesystem. Folders are handled correctly, too.

    This function takes three arguments:
        zipped    The ZipFile to extract the file from.
        file      The path in the ZipFile to the file to extract.
        dest      The path in the filesystem to the destination.

    """
    if '/' in dest:
        os.makedirs('/'.join(dest.split('/')[:-1]), exist_ok=True)
    if not file.endswith('/'):
        content = zipped.read(file)
        with open(dest, 'wb') as dest:
            dest.write(content)


def ask(question, want_yes=True, skip=False):
    """ Ask user for confirmation.

    This function accepts up to three parameters:
        question    The question to ask the user.
        want_yes    If Yes is the default alternative. Defaults to True.
        skip        If the confirmation should be skipped, and the question
                    printed just for illustration. Defaults to False.

    """
    print(question, end=' [Y/n]: ' if want_yes else ' [y/N]: ')
    if skip:
        print('y')
        return True
    else:
        answer = input()
        if len(answer) == 0:
            return want_yes
        else:
            return 'y' in answer.lower()



""" mcman main module. """

# Imports from the python library:
import argparse
import mcman
# Imports from mcman:
from mcman.frontend.plugins_command import PluginsCommand
from mcman.frontend.servers_command import ServersCommand


def negative(argument):
    """ Turn a number negative. """
    return -abs(int(argument))


def setup_server_commands(sub_parsers, parent):
    """ Setup the commands and subcommands for server. """
    # The parent parser for server and it's sub commands
    sub_parent = argparse.ArgumentParser(add_help=False, parents=[parent])

    # Base URL
    sub_parent.add_argument(
        '--base-url', metavar='base-url',
        default='http://spacegdn.totokaka.io/v1/',
        help='The base URL to use for SpaceGDN')

    # The server command parser
    parser = sub_parsers.add_parser(
        'server', aliases=['s'],
        help='Manage server jars',
        description='Download, identify and list Minecraft server jars.',
        parents=[sub_parent])
    parser.set_defaults(command=ServersCommand)
    # The server sub commands
    sub_parsers = parser.add_subparsers(title='subcommands')
    # servers, sub command of server
    servers_parser = sub_parsers.add_parser(
        'servers', aliases=['s'],
        help='List available servers.',
        description='List all server jars available for download.',
        parents=[sub_parent])
    servers_parser.set_defaults(subcommand='servers')
    # channels, sub command of server
    channels_parser = sub_parsers.add_parser(
        'channels', aliases=['c'],
        help='List channels for the specified server.',
        description='List all available channels for the server specified.',
        parents=[sub_parent])
    channels_parser.set_defaults(subcommand='channels')
    channels_parser.add_argument(
        'server', help='The server to get channels for.')
    # versions, sub command of server
    versions_parser = sub_parsers.add_parser(
        'versions', aliases=['v'],
        help='List versions for the specified server.',
        description='List all available versions for the server specified, '
                    + 'or only the versions in the specified channel.',
        parents=[sub_parent])
    versions_parser.set_defaults(subcommand='versions')
    versions_parser.add_argument(
        'server', help='The server to get versions for.')
    versions_parser.add_argument(
        'channel', nargs='?', help='The channel to get versions for.')
    # builds, sub command of server
    builds_parser = sub_parsers.add_parser(
        'builds', aliases=['b'],
        help='List builds for the specified server.',
        description='List all available builds for the server specified, '
                    + 'or only the ones in the specified channel, or version',
        parents=[sub_parent])
    builds_parser.set_defaults(subcommand='builds')
    builds_parser.add_argument(
        'server', help='The server to get builds for.')
    builds_parser.add_argument(
        'channel', nargs='?', help='The channel to get builds for.')
    builds_parser.add_argument(
        'version', nargs='?', help='The version to get builds for.')
    # download, sub command of server
    download_parser = sub_parsers.add_parser(
        'download', aliases=['d'],
        help='Download the newest version of the server.',
        description='Download the newest, the newest in the channel, or the '
                    + 'specified version of the jar.',
        parents=[sub_parent])
    download_parser.set_defaults(subcommand='download')
    download_parser.add_argument(
        'server', help='The server to download.')
    download_parser.add_argument(
        'channel', nargs='?', help='The channel to dowload from.')
    download_parser.add_argument(
        'version', nargs='?', help='The specific version to download.')
    download_parser.add_argument(
        'build', nargs='?', help='The specific build to download')
    download_parser.add_argument(
        '-o', '--output', help='Filename to download to')
    # identify, sub command of server
    identify_parser = sub_parsers.add_parser(
        'identify', aliases=['i'],
        help='Identify the server and version of the jar file.',
        description='Identifies the server, version and possibly channel of '
                    + 'the jar file specified.',
        parents=[sub_parent])
    identify_parser.set_defaults(subcommand='identify')
    identify_parser.add_argument(
        'jar', type=argparse.FileType('rb', 0),
        help='The jar file to identify.')

    return parser


def setup_plugin_commands(sub_parsers, parent):
    """ Setup the commands and subcommands for plugin. """
    # The parent parser for plugin and it's sub commands
    sub_parent = argparse.ArgumentParser(add_help=False, parents=[parent])

    # Base URL
    sub_parent.add_argument(
        '--base-url', default='http://api.bukget.org/3/',
        help='The base URL to use for BukGet')
    sub_parent.add_argument(
        '--server', default='bukkit',
        help='The server to get plugins for. This is sent to BukGet, '
             + 'and will affect what plugins you can download')
    sub_parent.add_argument(
        '--beta', action='store_const', dest='version', const='beta',
        help="Find latest beta version, instead of latest release for "
             + "plugins where a version isn't specified.")
    sub_parent.add_argument(
        '--alpha', action='store_const', dest='version', const='alpha',
        help="Find latest alpha version, instead of latest release for "
             + "plugins where a version isn't specified.")
    sub_parent.add_argument(
        '--latest', action='store_const', dest='version', const='latest',
        help="Find latest version, no matter type, instead of latest release "
             + "for plugins where a version isn't specified.")
    sub_parent.add_argument(
        '--no-resolve-dependencies', action='store_false',
        dest='resolve_dependencies', help='Do not resolve dependencies.')
    sub_parent.set_defaults(version='release')

    # The plugin command parser
    parser = sub_parsers.add_parser(
        'plugin', aliases=['p'],
        help='Manage plugins',
        description='Find, download and update plugins.',
        parents=[sub_parent])
    parser.set_defaults(command=PluginsCommand)

    # The plugin sub commands
    sub_parsers = parser.add_subparsers(title='subcommands')
    # search, sub command of plugin
    search_parser = sub_parsers.add_parser(
        'search', aliases=['s'],
        help='Search for a plugin',
        description='Search for a plugin using partial matching of the name.',
        parents=[sub_parent])
    search_parser.set_defaults(subcommand='search')
    search_parser.add_argument(
        'query', help='Search query.')
    # info, sub command of plugin
    info_parser = sub_parsers.add_parser(
        'info', aliases=['i'],
        help='Get info about a plugin(s)',
        description='Get info about one, or more plugins.',
        parents=[sub_parent])
    info_parser.set_defaults(subcommand='info')
    info_parser.add_argument(
        'plugins', metavar='plugin', type=str,
        help='Plugin(s) to get info for.')
    # download, sub command of plugin
    download_parser = sub_parsers.add_parser(
        'download', aliases=['d'],
        help='Download a plugin(s)',
        description='Download the specified plugin(s). A version can be '
                    + 'specified by appending "#<version>" to the plugin name',
        parents=[sub_parent])
    download_parser.set_defaults(subcommand='download')
    download_parser.add_argument(
        'plugins', metavar='plugin', type=str, nargs='+',
        help='Plugin(s) to download, and extract if they are zipped.')
    # update, sub command of plugin
    update_parser = sub_parsers.add_parser(
        'update', aliases=['u'],
        help='Update a plugin(s)',
        description='Update the specified plugins, or all.',
        parents=[sub_parent])
    update_parser.set_defaults(subcommand='update')
    update_parser.add_argument(
        'plugins', metavar='plugin', type=str, nargs='*',
        help='Plugin(s) update.')
    # list, sub command of plugin
    list_parser = sub_parsers.add_parser(
        'list', aliases=['l'],
        help='List installed plugins',
        description='List installed plugins, their versions, and the newest '
                    + 'version.',
        parents=[sub_parent])
    list_parser.set_defaults(subcommand='list')

    return parser


def main():
    """ Main function. """
    # Parent parser:
    parent = argparse.ArgumentParser(add_help=False)

    # Commands that can be used for all sub commands
    parent.add_argument(
        '--version',
        help='Show the version of mcman, then proceede normally',
        action='store_true', dest='show_version')
    parent.add_argument(
        '--user-agent',
        metavar='agent',
        type=str,
        default='mcman 0.1',
        help='Alternative user agent to report to BukGet and SpaceGDN')
    # Head and tail, they are mutually exclusive
    group = parent.add_mutually_exclusive_group()
    group.add_argument(
        '--head', '--size',
        metavar='size',
        type=int,
        dest='size',
        nargs='?',
        default=10,
        const=5,
        help='How many entries that should be displayed, from the top')
    group.add_argument(
        '--tail',
        metavar='size',
        type=negative,
        dest='size',
        nargs='?',
        default=10,
        const=-5,
        help='How many entries that should be displayed, from the bottom')
    group.add_argument(
        '--no-confirm',
        action='store_true',
        help='Do not wait for confirmation, just proceede.')

    # The top level command
    parser = argparse.ArgumentParser(
        description='Manage Minecraft server jars and plugins',
        epilog='Powered by BukGet and SpaceGDN',
        parents=[parent])

    # The sub commands, plugin and server
    sub_parsers = parser.add_subparsers(title='subcommands')

    server_parser = setup_server_commands(sub_parsers, parent)
    plugin_parser = setup_plugin_commands(sub_parsers, parent)

    args = parser.parse_args()
    if args.show_version:
        print('Version: {}'.format(mcman.__version__))
    # If there no command or subcommand is specified we print the help
    if not 'command' in args:
        parser.print_help()
        return
    if not 'subcommand' in args:
        if args.command is PluginsCommand:
            plugin_parser.print_help()
        elif args.command is ServersCommand:
            server_parser.print_help()
        return

    try:
        args.command(args)
    except KeyboardInterrupt:
        print()
        return



""" Status Handler. """


class StatusHandler(object):

    """ Status Handler.

    A status handler can register functions to run when a status is sent.

    """

    def __init__(self):
        """ Initialize. """
        self.statuses = dict()

    def register_handler(self, key, function):
        """ Register a handler. """
        self.statuses[key] = function

    def get_hook(self):
        """ Get the status hook. """
        def invoke(key, arguments):
            """ Invoke an handler. """
            if key in self.statuses:
                self.statuses[key](arguments)
            else:
                print('No handler found for key {}'.format(key))

        return invoke



""" The server command of mcman. """

from urllib.error import URLError
from mcman.backend import servers as backend
from mcman.backend import common as utils
from mcman.frontend.common import Command


class ServersCommand(Command):

    """ The server command of mcman. """

    def __init__(self, args):
        """ Parse command, and execute tasks. """
        Command.__init__(self)

        self.args = args

        backend.init(args.base_url, args.user_agent)

        self.register_subcommand('servers', self.servers)
        self.register_subcommand('channels', self.channels)
        self.register_subcommand('versions', self.versions)
        self.register_subcommand('builds', self.builds)
        self.register_subcommand('identify', self.identify)
        self.register_subcommand('download', self.download)

        self.invoke_subcommand(args.subcommand, (ValueError, URLError))

    def error(self, error):
        """ Print error from SpaceGDN. """
        self.p_sub('Error from SpaceGDN: {}'.format(error['message']))

    def result(self, head, results):
        """ Print results with head. """
        self.p_main(head)
        self.p_blank()
        if type(results) is str:
            self.p_sub(results)
        elif len(results) >= 1:
            self.p_sub(utils.list_names(results))
        else:
            self.p_sub('No results...')
        self.p_blank()

    def servers(self):
        """ List available servers. """
        self.p_main('Fetching server list from SpaceGDN')

        jars = backend.jars()

        if type(jars) is not list:
            self.error(jars)
        else:
            self.result('Available servers:', jars)

    def channels(self):
        """ List available channels for a server. """
        self.p_main('Fetching channel list from SpaceGDN')

        channels = backend.channels(self.args.server)

        if type(channels) is not list:
            self.error(channels)
        else:
            self.result('Available channels:', channels)

    def versions(self):
        """ List available versions of a server. """
        self.p_main('Fetching version list from SpaceGDN')

        versions = backend.versions(self.args.server, self.args.channel,
                                    self.args.size)

        if type(versions) is not list:
            self.error(versions)
        else:
            self.result('Available channels:', versions)

    def builds(self):
        """ List available builds for a server. """
        self.p_main('Fetching build list from SpaceGDN')

        builds = backend.builds(self.args.server, self.args.channel,
                                self.args.version, self.args.size)
        if type(builds) is not list:
            self.error(builds)
        else:
            self.result('Available builds:', builds)

    def identify(self):
        """ Identify what server a jar file is. """
        self.p_main('Calculating checksum of `{}`'.format(self.args.jar.name))

        checksum = utils.checksum_file(self.args.jar)
        self.args.jar.close()

        self.p_main('Finding build on SpaceGDN')

        build = backend.build_by_checksum(checksum)

        if build is None:
            self.p_main('Found no build on SpaceGDN with matching checksum')
            return

        server, channel, version, build = backend.get_roots(build)

        self.result('Found build:', '{} {} {} {}'.format(server, channel,
                                                         version, build))

    def download(self):
        """ Download a server. """
        self.p_main('Finding build on SpaceGDN')

        result = backend.get_builds(self.args.server, self.args.channel,
                                    self.args.version, self.args.build)

        if type(result) is not list:
            self.error(result)
            return

        if len(result) < 1:
            self.p_sub('Could not find any build')
            return

        channel, version, build = backend.find_latest_build(result)

        self.result('Found build:', '{} {} {} {}'.format(self.args.server,
                                                         channel, version,
                                                         build['build']))

        if utils.ask('Continue to download?', skip=self.args.no_confirm):
            utils.download(build['url'], destination=self.args.output,
                           checksum=build['checksum'], prefix=' '*4)
            self.p_blank()
            self.p_raw('Done!')



""" Common things for plugins_command and servers_command.

This module contains the classes 'UknownSubcommandException' and 'Command'.

The Command class is intended to be subclassed by all commands in mcman. The
class provides basic output methods, and handling of subcommands.

The UknownSubcommandException is an Exception that may be raised by certain
methods in the Command class.

"""


class UknownSubcommandException(BaseException):

    """ An exception raised by Command when a subcommand isn't found. """

    def __init__(self, subcommand):
        """ Initialize an exception.

        This initializer requires one parameter:
            subcommand    The subcommand that was not found.

        """
        BaseException.__init__('Could not find subcommand {}'
                               .format(subcommand))
        self.subcommand = subcommand


class Command(object):

    """ A common Command class.

    Intended for use by the servers and plugins commands. It provieds methods
    to output things nicely, and registering and invoking of sub commands.

    """

    def __init__(self, prefix=' :: ', printer=print):
        """ Initialize this command.

        This method sets the prefix, and creates a matching empty prefix,
        that just contains spaces.

        Up to two parameters are accepted:
            prefix=' :: '    The prefix.
            printer=print    The method used to build. This must support some
                             of the same parameters as python's built in print
                             most notably: A named 'sep', passing of multiple
                             strings.

        """
        self.printer = printer

        self.f_prefix = prefix
        self.e_prefix = ' ' * len(prefix)

        self.subcommands = dict()

    def p_main(self, message, frmt=None):
        """ Print a main line.

        A main line has the prefix before it.
        This method takes up to to arguments:
            message      The message to print.
            frmt=None    A format to format the message by.
                         This is passed to message.format()

        """
        self.p_raw(self.f_prefix, message.format(frmt), sep='')

    def p_sub(self, message, frmt=None):
        """ Print a sub line.

        A sub line is the 'child' of a main line. It is intended with the same
        lenght as the prefix, but with tabs.
        This method takes up to to arguments:
            message      The message to print.
            frmt=None    A format to format the message by.
                         This is passed to message.format()

        """
        self.p_raw(self.e_prefix, message.format(frmt), sep='')

    def p_blank(self):
        """ Print a blank line. """
        self.p_raw()

    def p_raw(self, *args, **kwargs):
        """ Print directly to the printer. """
        self.printer(*args, **kwargs)

    def register_subcommand(self, name, method):
        """ Register a subcommand.

        It will be stored by the name, and it can later be invoked by doing
        'self.invoke_subcommand(name)'.

        This method accepts two parameters:
            name      The name of the subcommand.
            method    The method that contains this subcommand.
                      The  method may only accept the 'self' parameter.

        """
        self.subcommands[name] = method

    def invoke_subcommand(self, name, exceptions):
        """ Invoke a subcommand.

        This method will invoke a previously registered subcommand, by the
        name. It will also catch the specified tuple of exceptions and just
        display them as errors, without a stack trace.

        two parameters are accepted:
            name          The name of the subcommand to invoke.
            exceptions    A tuple of exceptions to be catched

        Exceptions raised:
            UknownSubcommandException    If the subcommand wasn't found or
                                         registered.

        Additionally any exception may be raised by the subcommand invoked.

        """
        if not name in self.subcommands:
            raise UknownSubcommandException(name)
        else:
            try:
                self.subcommands[name]()
            except UknownSubcommandException as err:
                self.p_main('Could not find subcommand {}'.format(
                    err.subcommand))
            except exceptions as err:
                self.p_sub('Error: {}'.format(str(err)))



""" The plugin command of mcman.

This module is the home of the front end part of the command. This menas that
as little as possible logic should go here. The logic is placed the the plugins
module in the backend package.

"""

from urllib.error import URLError
from mcman.status_handler import StatusHandler
from mcman.backend import plugins as backend
from mcman.backend import common as utils
from mcman.frontend.common import Command


class PluginsCommand(Command):

    """ The plugin command of mcman. """

    def __init__(self, args):
        """ Parse command, and execute tasks. """
        Command.__init__(self)

        self.server = args.server
        self.args = args

        backend.init(args.base_url, args.user_agent)
        backend.VERSION = self.args.version

        self.register_subcommand('search', self.search)
        self.register_subcommand('info', self.info)
        self.register_subcommand('list', self.list)
        self.register_subcommand('download', self.download)
        self.register_subcommand('update', self.update)

        self.invoke_subcommand(args.subcommand, (ValueError, URLError))

    def search(self):
        """ Search for plugins. """
        query = self.args.query

        self.p_main('Searching for `{}` through BukGet'.format(query))

        results = [(result, '[{}]{}'.format(result['slug'],
                                            result['plugin_name']))
                   for result in backend.search(query, self.args.size)]

        # Find lenght of longest slug + name
        max_len = max([len(p[1]) for p in results])

        # Create the format
        frmt = '{{:<{}}} - {{}}'.format(max_len)

        self.p_main('Results:')
        self.p_blank()

        for i in range(min(len(results), abs(self.args.size))):
            plugin = results[i]
            line = frmt.format(plugin[1], plugin[0]['description'].strip())
            self.p_sub(line)

        self.p_blank()

    def info(self):
        """ Display info about a plugin. """
        query = self.args.plugins
        self.p_main('Finding `{}` on BukGet'.format(query))

        plugin = backend.info(self.server, query)
        if plugin is None:
            self.p_sub('Could not find `{}`'.format(query))
            return

        self.p_main('Found {}:'.format(plugin['plugin_name']))
        self.p_blank()

        website = plugin['website']
        if len(website) == 0:
            website = plugin['dbo_page']
        authors = utils.list_names(plugin['authors'])
        categories = utils.list_names(plugin['categories'])
        stage = plugin['stage']
        name = plugin['plugin_name']
        desciption = plugin['description'].strip()

        self.p_sub('Name:        {}', name)
        self.p_sub('Description: {}', desciption)
        self.p_sub('Authors:     {}', authors)
        self.p_sub('Website:     {}', website)
        self.p_sub('Categories:  {}', categories)
        self.p_sub('Stage:       {}', stage)

        self.p_blank()
        self.p_sub('Versions:')

        versions = plugin['versions']

        # Sorting
        if self.args.size >= 0:
            versions = versions[:min(self.args.size, len(versions))]
        else:
            versions = versions[max(self.args.size, -len(versions)):]

        # Formatting
        max_len = max([len(v['version']) for v in versions])
        frmt = '    {{:>{}}} - {{}}'.format(max_len)

        for version in versions:
            line = frmt.format(version['version'],
                               utils.list_names(version['game_versions']))
            self.p_sub(line)

        self.p_blank()

    def list(self):
        """ List installed plugins. """
        self.p_main('Finding installed plugins')

        plugins = backend.list_plugins()

        if len(plugins) == 0:
            self.p_sub('Found no plugins')
            return

        self.p_main('Looking up plugins on BukGet')

        new_versions = dict()
        for yielded in backend.find_newest_versions(plugins, self.server):
            if type(yielded) is tuple:
                slug, version = yielded
                new_versions[slug] = version
            elif type(yielded) is str:
                self.p_sub('Could not find {} on BukGet'.format(yielded))

        self.p_main('Installed plugins:')
        self.p_blank()

        max_name_len = max([len('[{}]{}'.format(p[3], p[2])) for p in plugins])
        max_ver_len = max([len(p[1]) for p in plugins])
        frmt = '{{:<{}}} - {{:<{}}}'.format(max_name_len, max_ver_len)

        for slug, version, name, jar in plugins:
            line = frmt.format('[{}]{}'.format(jar, name), version)
            if slug in new_versions:
                new_version = new_versions[slug]
                line += '  -- Out of date, newest version: {}'.format(
                    new_version)
            self.p_sub(line)

        self.p_blank()

    def download(self):
        """ Download plugins. """
        self.p_main('Finding plugins on BukGet')

        status_handler = StatusHandler()
        status_handler.register_handler(1, lambda arguments:
                                        self.p_sub('Could not find {}'.format(
                                            arguments[0])))
        to_install, versions = backend.find_plugins(self.args.server,
                                                    self.args.plugins,
                                                    status_handler.get_hook())

        self.p_main('Resolving dependencies')

        status_handler = StatusHandler()
        one_three = lambda arguments: \
            self.p_sub('Could not find `{}`'.format(arguments))
        two = lambda arguments: \
            self.p_sub('Could not find version `{}` of `{}`'.format(
                arguments[0], arguments[1]))
        status_handler.register_handler(1, one_three)
        status_handler.register_handler(2, two)
        status_handler.register_handler(3, one_three)
        to_install = backend.dependencies(self.args.server, to_install,
                                          versions, status_handler.get_hook())

        if len(to_install) < 1:
            self.p_main('Found no plugins!')
            return

        self.p_main(
            'Resolving versions, and checking allready installed plugins')

        status_handler = StatusHandler()
        one = lambda argument: \
            self.p_sub('Could not find plugin `{}` again'.format(argument))
        two = lambda argument: \
            self.p_sub('Could not find version of `{}` again'.format(
                argument))
        three = lambda argument: \
            self.p_sub('{} was allready installed.'.format(
                argument))
        four = lambda argument: \
            self.p_sub('{} is allready installed, but out of date'.format(
                argument))
        status_handler.register_handler(1, one)
        status_handler.register_handler(2, two)
        status_handler.register_handler(3, three)
        status_handler.register_handler(4, four)

        plugins = backend.find_updates(self.server, to_install, versions,
                                       status_handler.get_hook())

        if len(plugins) < 1:
            self.p_main('No plugins left to install')
            return

        self.p_main('Plugins to install:')
        self.p_blank()
        self.p_sub(utils.list_names([p['plugin_name'] + '#'
                                    + p['versions'][0]['version']
                                    for p in plugins]))
        self.p_blank()

        backend.download('Continue to download?',
                         '({{part:>{}}}/{{total}}) '.format(
                             len(str(len(plugins)))),
                         plugins, self.args.no_confirm)

        self.p_blank()
        self.p_raw('Done!')

    def update(self):
        """ Update installed plugins. """
        self.p_main('Finding installed plugins')

        installed = backend.list_plugins()

        self.p_main('Looking up versions on BukGet')

        to_update = list()
        for i in installed:
            plugin, i_version, name = i[0], i[1], i[2]
            u_plugin = backend.download_details(self.server, plugin,
                                                self.args.version)
            if u_plugin is None or len(u_plugin['versions']) < 1:
                self.p_sub('Could not find {} on BukGet!'.format(name))
                continue

            u_version = u_plugin['versions'][0]['version']
            if u_version > i_version:
                to_update.append(u_plugin)

        self.p_main('Plugins to update:')
        self.p_blank()

        if len(to_update) < 1:
            self.p_sub('All plugins are up to date!')
            self.p_blank()
            return

        self.p_sub(utils.list_names([p['plugin_name'] + '#'
                                     + p['versions'][0]['version']
                                     for p in to_update]))
        self.p_blank()

        if utils.ask('Continue to update?', skip=self.args.no_confirm):
            prefix_format = '({{part:>{}}}/{{total}}) '.format(
                len(str(len(to_update))))

            for i in range(len(to_update)):
                plugin = to_update[i]
                prefix = prefix_format.format(total=len(to_update), part=i+1)
                backend.download_plugin(plugin, prefix)
            self.p_blank()
            self.p_raw('Done!')



""" The backend for the mcman servers command. """
import spacegdn


def init(base, user_agent):
    """ Initialize this module.

    This function will just set the base url and user agent for SpaceGDN.

    """
    spacegdn.BASE = base
    spacegdn.USER_AGENT = user_agent


def jars():
    """ List jars.

    A list of the jar names is returned if the query was successful, else the
    error dictionary from SpaceGDN is returned.

    """
    result = spacegdn.jars()

    if type(result) is not list:
        return result

    return [jar['name'] for jar in result]


def channels(server):
    """ List channels.

    A list of the channel names is returned if the query was successful, else
    the error dictionary from SpaceGDN is returned.

    """
    server = spacegdn.get_id(jar=server)
    result = spacegdn.channels(jar=server)

    if type(result) is not list:
        return result

    return [channel['name'] for channel in result]


def versions(server, channel, size):
    """ List versions.

    A list of the version names is returned if the query was successful, else
    the error dictionary from SpaceGDN is returned.

    """
    server = spacegdn.get_id(jar=server)
    if channel is not None:
        channel = spacegdn.get_id(jar=server, channel=channel)
    result = spacegdn.versions(jar=server, channel=channel)

    if type(result) is not list:
        return result

    # Sort and limit the results
    result = list(reversed(sorted([version['version'] for version
                                   in result])))
    if size >= 0:
        result = result[:min(size, len(result))]
    else:
        result = result[max(size, -len(result)):]

    return result


def builds(server, channel, version, size):
    """ List builds.

    A list of the build numbers is returned if the query was successful, else
    the error dictionary from SpaceGDN is returned.

    """
    server = spacegdn.get_id(jar=server)
    if channel is not None:
        channel = spacegdn.get_id(jar=server, channel=channel)
    if version is not None:
        version = spacegdn.get_id(jar=server, channel=channel, version=version)
    result = spacegdn.builds(jar=server, channel=channel, version=version)

    if type(result) is not list:
        return result

    # Sort and limit the results
    result = list(reversed(sorted([str(build['build']) for build
                                   in result])))
    if size >= 0:
        result = result[:min(size, len(result))]
    else:
        result = result[max(size, -len(result)):]

    return result


def get_builds(server, channel, version, build):
    """ Get the build.

    Any argument might be None.

    A list of matching builds is returned on success, or the error dictionary
    as returned by SpaceGDN.

    """
    server = spacegdn.get_id(jar=server)
    if channel is not None:
        channel = spacegdn.get_id(jar=server, channel=channel)
    if version is not None:
        version = spacegdn.get_id(jar=server, channel=channel,
                                  version=version)
    if build is not None:
        build = spacegdn.get_id(jar=server, channel=channel,
                                version=version,
                                build=build)

    return spacegdn.builds(jar=server, channel=channel,
                           version=version, build=build)


def build_by_checksum(checksum):
    """ Find build by checksum.

    None is returned if the build was not found.

    """
    result = spacegdn.builds(where='build.checksum.eq.{}'.format(checksum))
    if len(result) < 1:
        return None
    return result[0]


def get_roots(build):
    """ Get the roots(server, channel, version and build) of this build.

    The returned value is a tuple of the server name, channel name, version
    name and build number.

    """
    server = spacegdn.jars(build['jar_id'])[0]['name']
    channel = spacegdn.channels(build['jar_id'],
                                build['channel_id'])[0]['name']
    version = spacegdn.versions(build['jar_id'], build['channel_id'],
                                build['version_id'])[0]['version']
    build = build['build']

    return server, channel, version, build


def find_latest_build(build_list):
    """ Find the latest build in a list of builds. """
    build_list.sort(key=lambda build: build['build'], reverse=True)
    build = build_list[0]

    channel = spacegdn.channels(channel=build['channel_id'])[0]['name']
    version = spacegdn.versions(version=build['version_id'])[0]['version']

    return channel, version, build



""" The backend for the mcman command. """

import os
import bukget
import yaml
from zipfile import ZipFile
from mcman.backend import common as utils


VERSION = 'release'


def init(base, user_agent):
    """ Initialize the module.

    This function just sets the base url and user agent for BukGet, and it
    sets the default version type.

    """
    bukget.BASE = base
    bukget.USER_AGENT = user_agent


def search(query, size):
    """ Search for plugins.

    This function will search for plugins on bukget by their slug and name.
    The results are ordered by a score calculated with their levenshtein edit
    distance from the search query and their popularity.

    Two parameters are required:
        query    The query to search for.
        size     The size of the returned results. It can be negative, and in
                 that case the bottom of the results are returned.

    A list with the plugin dicts from BukGet is returned. It is sorted by the
    score system mentioned earlier.

    """
    sorting = ('-' if size >= 0 else '')+'popularity.monthly'
    search_results = bukget.search(
        {
            'field': 'plugin_name',
            'action': 'like',
            'value': query
        },
        sort=sorting,
        fields='slug,plugin_name,description,popularity.monthly',
        size=abs(size))
    search_results += bukget.search(
        {
            'field': 'slug',
            'action': 'like',
            'value': query
        },
        sort=sorting,
        fields='slug,plugin_name,description,popularity.monthly',
        size=abs(size))

    new_results = list()
    for result in search_results:
        for result2 in new_results:
            if result2['slug'] == result['slug']:
                break
        else:
            new_results.append(result)
    search_results = new_results

    # Calculate scores
    results = list()
    for plugin in search_results:
        if len(plugin) == 0:
            continue

        score = plugin['popularity']['monthly'] - utils.levenshtein(
            query, plugin['plugin_name'])

        results.append((score, plugin))

    results.sort(key=lambda x: x[0], reverse=True)

    return [result[1] for result in results]


def info(server, name):
    """ Find the plugin by name, and return certain information.

    The returned information is the information that is needed for the info
    command.

    The plugin dictionary as returned by pyBukGet is returned, or None if the
    plugin was not found.

    """
    slug = bukget.find_by_name(server, name)
    if slug is None:
        return None

    plugin = bukget.plugin_details(server, slug,
                                   fields='website,dbo_page,'
                                          + 'description,'
                                          + 'versions.type,'
                                          + 'versions.game_versions,'
                                          + 'versions.version,'
                                          + 'plugin_name,server,'
                                          + 'authors,categories,'
                                          + 'stage,slug')

    return plugin


def find_newest_versions(plugins, server):
    """ Find newest versions for the plugins.

    This is a generator. When a plugin in the plugins list has an update it is
    yielded as a tuple with the plugin slug and the newest version. If the
    plugin is not found at BukGet, the plugin name is yielded alone. If a
    plugin does not have any updates, nothing is yielded.

    Three parameters are required:
        plugins    A list of plugins, as returned by list_plugins()
        server     The server the plugins are for.

    """
    for plugin in plugins:
        slug, version, name = plugin[0], plugin[1], plugin[2]
        plugin = bukget.plugin_details(server, slug,
                                       VERSION,
                                       fields='versions.version')
        if plugin is None:
            yield name
            continue
        b_version = plugin['versions'][0]['version']
        if b_version > version:
            yield slug, b_version


def download_plugin(plugin, prefix=""):
    """ Download plugin.

    This function takes two parameters. The first is the plugin. The plugin
    must be a dictionary returned by bukget with at least the following fields:
        versions.download, plugin_name, versions.md5, versions.filename
    The second parameter is a prefix to print before each output line,
    typically a counter on which download this it:
        ( 5/20)

    """
    target_folder = find_plugins_folder() + '/'

    url = plugin['versions'][0]['download']
    if ' ' in plugin['plugin_name']:
        filename = ''.join([x.capitalize() for x
                            in plugin['plugin_name'].split(' ')])
    else:
        filename = plugin['plugin_name']
    md5 = plugin['versions'][0]['md5']
    suffix = plugin['versions'][0]['filename'].split('.')[-1:][0]
    full_name = target_folder + filename + '.' + suffix

    utils.download(url, destination=full_name,
                   checksum=md5, prefix=prefix, display_name=filename)

    if suffix == 'zip':
        print(' '*len(prefix) + 'Unzipping...', end=' ')
        unzip_plugin(full_name, target_folder)
        os.remove(full_name)
        print('Success')


def unzip_plugin(target_file, target_folder):
    """ Unzip a plugin that is packaged in a zip. """
    with ZipFile(target_file, 'r') as zipped:
        folders = [x for x in zipped.namelist()
                   if x.endswith('/')]
        jars = [x for x in zipped.namelist()
                if x.endswith('.jar')]
        strip_folder = False
        if len(jars) == 0 and len(folders) > 0:
            folder = folders[0]
            jars = [x for x in zipped.namelist()
                    if x.endswith('.jar')
                    and x.startswith(folder)]
            strip_folder = folder

        for jar in jars:
            utils.extract_file(zipped, jar,
                               target_folder + (
                                   jar.split(strip_folder, 1)[1]
                                   if strip_folder else jar))


def list_plugins():
    """ List installed plugins.

    Returns a set of four-tuples containing the slug, version, plugin_name and
    path to the jar, in that order.

    """
    plugins = set()

    folder = find_plugins_folder()
    jars = [folder + '/' + f for f in os.listdir(folder)
            if os.path.isfile(folder + '/' + f) and f.endswith('.jar')]
    for jar in jars:
        # We first search for a plugin with a published version with matching
        # checksum to ours.
        slug = None
        plugin_name = None
        checksum = utils.checksum_file(jar)
        result = bukget.search({'field':  'versions.md5',
                                'action': '=',
                                'value':  checksum},
                               fields='slug,versions.version,plugin_name')
        if len(result) > 0:
            slug = result[0]['slug']
            plugin_name = result[0]['plugin_name']
        # Then we search by main class
        with ZipFile(jar, 'r') as zipped:
            # If there is no plugin.yml in it
            # we ignore it
            if not 'plugin.yml' in zipped.namelist():
                continue
            yml = yaml.load(zipped.read('plugin.yml').decode())
            if slug is None:
                plugin_name = yml['name']
                main = yml['main']
                result = bukget.search({'field':  'main',
                                        'action': '=',
                                        'value':  main},
                                       fields='slug,versions.version')
                if len(result) > 0:
                    slug = result[0]['slug']
                else:
                    continue
            version = str(yml['version'])
            plugins.add((slug, version, plugin_name, jar))

    return plugins


def find_plugins_folder():
    """ Find the plugins folder.

    This will return the relative path to the plugins folder.
    Currently either '.' or 'plugins' is returend.

    """
    if 'plugins' in os.listdir('.'):
        return 'plugins'
    return '.'


def dependencies(server, plugins, versions, status_hook):
    """ Get list of dependencies to install by list of plugins.

    This function will return a list containing all the plugins in the supplied
    list of plugins plus all their dependencies.

    Parameters:
        server         The server.
        plugins        A list of plugins.
        versions       A dictionary mapping from plugins to versions.
        status_hook    A status hook, see the docs on resolve_dependencies.

    """
    new_stack = list()
    for plugin in plugins:
        resolve_dependencies(server,
                             plugin,
                             status_hook,
                             versions[plugin] if plugin in versions else None,
                             new_stack)
    return new_stack


def resolve_dependencies(server, plugin_name, status_hook,
                         version=None, stack=None):
    """ Resolve all dependencies of plugin.

    This function will recursively get a list of all dependencies of the
    plugin and the plugin itself. This means that the returned list
    contains all the plugins needed to install.

    A ValueError will be raised if a plugin isn't found. This can either be
    the plugin itself, or any of it's dependencies.

    Parameters:
       server          The server the plugin is for.
       plugin_name     The name of the plugin.
       status_hook     A status hook, explained later.
       version         An optional specific version of the plugin.
       dependencies    plugins that will allready be installed.

    The status hook should be a function that accepts an integer and a tuple.
    The integer says what status it is, and the tuple contains the values.
    Here is a list of the statuses:
        1 - Plugin not found on BukGet. Values: plugin name
        2 - Version not found on BukGet. Values: plugin name, version that was
            tried.
        3 - Dependency not found on BukGet. Values: dependency name

    """
    if version is None:
        version = VERSION
    if stack is None:
        stack = list()

    plugin = bukget.find_by_name(server, plugin_name)
    if plugin is None:
        status_hook(1, plugin_name)
        return stack

    plugin = bukget.plugin_details(server, plugin, version,
                                   fields='slug,'
                                          + 'versions.hard_dependencies,'
                                          + 'versions.version')

    if plugin['slug'] in stack:
        return stack
    if len(plugin['versions']) < 1:
        status_hook(2, (plugin_name, version))
        return stack

    stack.append(plugin['slug'])

    for dep in plugin['versions'][0]['hard_dependencies']:
        slug = bukget.find_by_name(server, dep)
        if slug is None:
            status_hook(3, dep)
            continue
        if slug not in stack:
            resolve_dependencies(server, slug, status_hook, stack=stack)

    return stack


def find_plugins(server, queries, status_hook):
    """ Find plugins.

    Parameters:
        server         Server type the plugins are for
        queryies       Plugins to try and find.
        status_hook    A status hook, see below.

    One of the parameters is a status hook. It should be a function that
    accepts an integer and a tuple. The integer is the status id, the tuple
    contains the values.

    Statuses:
        1 - Plugin not found. Values: plugin

    A tuple of a list and dict is returned. the list contains the plugins to
    install, while the versions dict contains specific versions.

    """
    to_install = list()
    versions = dict()

    for query in queries:
        plugin_version = query.rsplit('#', 1)
        plugin = plugin_version[0]
        version = ''
        slug = bukget.find_by_name(server, plugin)
        if slug is None:
            status_hook(1, plugin)
            continue
        if len(plugin_version) > 1:
            version = plugin_version[1]
            versions[slug] = version
        to_install.append(slug)

    return to_install, versions


def download_details(server, plugin, version):
    """ Get details required for download.

    This function is just a wrapper around bukget, that will get the details
    about a plugin that is required when downloading it.

    """
    return bukget.plugin_details(server, plugin, version,
                                 fields='slug,plugin_name,versions.version,'
                                        + 'versions.md5,versions.download,'
                                        + 'versions.type,'
                                        + 'versions.filename,'
                                        + 'versions.hard_dependencies,'
                                        + 'versions.soft_dependencies')


def download(question, frmt, plugins, skip=False):
    """ Download plugins.

    This function does the questioning, and installing of each plugin.

    Parameters:
        question    The question to ask the user.
        frmt        The format to format the prefix with. Must support two
                    parameters: total and part.
        plugins     A list of the plugins to install.
        skip        Whether to skip the confirmation.

    """
    if utils.ask(question, skip=skip):
        for i in range(len(plugins)):
            plugin = plugins[i]
            prefix = frmt.format(total=len(plugins), part=i+1)
            download_plugin(plugin, prefix)


def find_updates(server, to_install, versions, status_hook):
    """ Find updates for plugins.

    This function will also check if the plugins are allready installed.

    Parameters:
        server         The server
        to_install     A list over plugins that should be installed
        versions       A dict with optional specific versions for plugins
        status_hook    A status hook

    """
    plugins = list()
    installed = list_plugins()

    for slug in to_install:
        plugin = download_details(server, slug,
                                  versions[slug]
                                  if slug in versions
                                  else VERSION)

        if plugin is None:
            status_hook(1, slug)
            continue
        elif len(plugin['versions']) < 1:
            status_hook(2, slug)
            continue

        for i in installed:
            if i[0] == slug and i[1] >= plugin['versions'][0]['version']:
                status_hook(3, plugin['plugin_name'])
                break
            elif i[0] == slug:
                status_hook(4, plugin['plugin_name'])
        else:
            plugins.append(plugin)

    return plugins



""" Setup file for PlugMan. """
from setuptools import setup
import mcman

setup(
    name='mc-man',
    version=mcman.__version__,
    author=mcman.__author__,
    author_email='mail@totokaka.io',
    description='A Minecraft server jar and plugins manager',
    url='https://github.com/CubeEngineDev/mc-man',
    packages=['mcman', 'mcman.frontend', 'mcman.backend'],
    scripts=["bin/mcman"],
    install_requires=['PyYAML>=3.10',
                      'pyBukGet>=2.3',
                      'pySpaceGDN>=0.2'],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities'
    ],
)
